/*
 * Filter
 * Copyright (C) 2011 by Apaq
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package dk.apaq.filter.jpa;

import dk.apaq.filter.jpa.interpreters.Interpreter;
import dk.apaq.filter.jpa.interpreters.LikeFilterInterpreter;
import dk.apaq.filter.Filter;
import dk.apaq.filter.core.AndFilter;
import dk.apaq.filter.core.CompareFilter;
import dk.apaq.filter.core.ContainsFilter;
import dk.apaq.filter.core.IsNullFilter;
import dk.apaq.filter.core.LikeFilter;
import dk.apaq.filter.core.NotFilter;
import dk.apaq.filter.core.OrFilter;
import dk.apaq.filter.jpa.interpreters.CompareFilterInterpreter;
import dk.apaq.filter.jpa.interpreters.ContainsFilterInterpreter;
import dk.apaq.filter.jpa.interpreters.IsNullFilterInterpreter;
import dk.apaq.filter.limit.Limit;
import dk.apaq.filter.sort.SortDirection;
import dk.apaq.filter.sort.Sorter;
import dk.apaq.filter.sort.SorterEntry;
import java.util.HashMap;
import java.util.Map;
import javax.persistence.EntityManager;
import javax.persistence.Query;

/**
 *  Translates Filter, Sorters and Limits to JPA Query.
 */
public class FilterTranslatorForJPA {

    private static final Map<Class<? extends Filter>, Interpreter<? extends Filter>> interpreterMap = new HashMap<Class<? extends Filter>, Interpreter<? extends Filter>>();

    static {
        interpreterMap.put(LikeFilter.class, new LikeFilterInterpreter());
        interpreterMap.put(ContainsFilter.class, new ContainsFilterInterpreter());
        interpreterMap.put(CompareFilter.class, new CompareFilterInterpreter());
        interpreterMap.put(IsNullFilter.class, new IsNullFilterInterpreter());
    }

    public static void registerInterpreter(Class<? extends Filter> clazz, Interpreter<? extends Filter> interpreter) {
        interpreterMap.put(clazz, interpreter);
    }

    public static Query translate(EntityManager entityManager, Class entityClass, Filter filter, Sorter sorter, Limit limit) {
        return translate(entityManager, null, entityClass, filter, sorter, limit);
    }

    public static Query translate(EntityManager entityManager, String[] fields, Class entityClass, Filter filter, Sorter sorter, Limit limit) {

        WhereClause whereClause = new WhereClause();
        boolean clauseAppended = buildWhereClause(whereClause, filter);

        StringBuilder jpqlBuilder = new StringBuilder();
        jpqlBuilder.append("select ");
        if (fields == null || fields.length == 0) {
            jpqlBuilder.append("e");
        } else {
            boolean firstField = true;
            for (String field : fields) {
                if (!firstField) {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append("e.");
                jpqlBuilder.append(field);
                firstField = false;
            }
        }

        jpqlBuilder.append(" from ");
        jpqlBuilder.append(entityClass.getSimpleName());
        jpqlBuilder.append(" e");

        if (clauseAppended) {
            jpqlBuilder.append(" WHERE ");
            jpqlBuilder.append(whereClause.getStatement());
        }

        if (sorter != null && !sorter.getSorterEntries().isEmpty()) {
            boolean firstOrderClause = true;
            for (SorterEntry entry : sorter.getSorterEntries()) {
                if (firstOrderClause) {
                    jpqlBuilder.append(" ORDER BY ");
                } else {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append(entry.getPropertyId());
                jpqlBuilder.append(entry.getDirection() == SortDirection.Ascending ? " ASC" : " DESC");
                firstOrderClause = false;
            }
        }

        Query q = entityManager.createQuery(jpqlBuilder.toString());
        for (Map.Entry<String, Object> paramEntry : whereClause.getParameters().entrySet()) {
            q.setParameter(paramEntry.getKey(), paramEntry.getValue());
        }

        if (limit != null) {
            q.setFirstResult(limit.getOffset());
            q.setMaxResults(limit.getCount());
        }

        return q;

    }

    private static boolean buildWhereClause(WhereClause clause, Filter filter) {

        if (filter == null) {
            return false;
        }

        //First try an interpreter
        Interpreter interpreter = interpreterMap.get(filter.getClass());
        if (interpreter != null) {
            return interpreter.interpret(clause, filter);
        }


        //..else try our special handlers for filtercontainers.
        if (filter instanceof AndFilter) {
            AndFilter af = (AndFilter) filter;

            clause.appendStatement("(");

            boolean hasAddedFilter = false;
            for (Filter currentFilter : af.getFilters()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentFilter);

                if (statementAppended) {
                    if (hasAddedFilter) {
                        clause.insertStatement(startIndex, " AND ");
                    }
                    hasAddedFilter = true;
                }
            }

            clause.appendStatement(")");

            return hasAddedFilter;
        }

        if (filter instanceof OrFilter) {
            OrFilter of = (OrFilter) filter;
            boolean hasAddedFilter = false;

            clause.appendStatement("(");

            for (Filter currentFilter : of.getFilters()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentFilter);

                if (statementAppended) {
                    if (hasAddedFilter) {
                        clause.insertStatement(startIndex, " OR ");
                    }
                    hasAddedFilter = true;
                }
            }

            clause.appendStatement(")");

            return hasAddedFilter;
        }

        if (filter instanceof NotFilter) {
            NotFilter af = (NotFilter) filter;

            //Special care of IsNull filter becauseof the special "IS [NOT] NULL" syntax
            if (af.getFilterElement() instanceof IsNullFilter) {
                String propId = ((IsNullFilter) af.getFilterElement()).getPropertyId();
                clause.appendStatement("e." + propId + " IS NOT NULL");
                return true;
            }

            int startIndex = clause.getStatementLength();
            boolean statementAppended = buildWhereClause(clause, af.getFilterElement());
            if (statementAppended) {
                clause.insertStatement(startIndex, "NOT (");
                clause.appendStatement(")");
                return true;
            } else {
                return false;
            }
        }

        return false;
    }
}
